import io
import os.path
from pathlib import Path

import geojson
from fs import open_fs
from geojson import FeatureCollection

from plateaukit.logger import logger
from ._single import geojson_from_gml_single


def geojson_from_gml_serial_with_quit(
    infiles,
    outfile,
    codelist_infiles=None,
    types=None,
    include_type=False,
    zipfile=None,
    task_id=None,
    _quit=None,
    _progress=None,
    **kwargs,
):
    """Generate GeoJSON from multiple CityGML files."""
    features = []

    # Load codelists
    codelist_file_map = None

    if zipfile is not None:
        zip_fs = open_fs(f"zip://{zipfile}")

    if codelist_infiles and infiles:
        base_path = Path(infiles[0]).parent  # TODO: Fix this
        codelist_file_map = dict()

        for codelist_infile in codelist_infiles:
            if zipfile is not None:
                with zip_fs.openbin(codelist_infile, "r") as f:
                    relative_path = os.path.relpath(codelist_infile, base_path)
                    codelist_file_map[relative_path] = io.BytesIO(f.read())
            else:
                with open(codelist_infile, "rb") as f:
                    relative_path = os.path.relpath(codelist_infile, base_path)
                    codelist_file_map[relative_path] = io.BytesIO(f.read())

    total = len(infiles) + 1  # + 1 for geojson.dump

    for i, infile in enumerate(infiles):
        if task_id is not None and _progress is not None:
            _progress[task_id] = {"progress": i + 1, "total": total}

        if _quit and _quit.is_set():
            return

        logger.debug(f"infile: {infile}")

        if zipfile:
            with zip_fs.openbin(infile, "r") as f:
                collection = geojson_from_gml_single(
                    f,
                    types=types,
                    codelist_file_map=codelist_file_map,
                    include_type=include_type,
                    **kwargs,
                )
        else:
            with open(infile, "r") as f:
                collection = geojson_from_gml_single(
                    f,
                    types=types,
                    codelist_file_map=codelist_file_map,
                    include_type=include_type,
                    **kwargs,
                )
                # TODO: fix

        try:
            features.extend(collection["features"])
        except Exception as err:
            logger.debug(err)

    if zipfile:
        zip_fs.close()

    collection = FeatureCollection(features)

    # NOTE: Precision https://github.com/jazzband/geojson#default-and-custom-precision
    with open(outfile, "w") as f:
        geojson.dump(collection, f, ensure_ascii=False, separators=(",", ":"))

    # Complete progress
    _progress[task_id] = {"progress": total, "total": total}



import glob
import tempfile
from pathlib import Path

from rich import get_console

from plateaukit.config import Config
from plateaukit.dataset import load_dataset


def prebuild(dataset_id: str, *, split: int = 10, simple_output=False) -> None:
    """Prebuild PLATEAU datasets."""

    import geopandas as gpd
    import pandas as pd
    from pyogrio import read_dataframe, write_dataframe

    console = get_console()

    if not dataset_id:
        raise Exception("Missing argument: dataset_id")

    config = Config()
    record = config.datasets.get(dataset_id)
    # print(dataset_id, record)

    # TODO: All types
    types = ["bldg"]

    with tempfile.TemporaryDirectory() as tdir:
        for type in types:
            outfile = Path(tdir, f"{dataset_id}.{type}.geojson")

            if dataset_id:
                dataset = load_dataset(dataset_id)
                dataset.to_geojson(
                    outfile,
                    types=types,
                    altitude=False,  # TODO: Check this
                    include_type=True,
                    split=split,
                    progress={"description": "Generating GeoJSON files..."},
                    simple_output=simple_output,
                )
            else:
                raise NotImplementedError()
                # generators.geojson.geojson_from_citygml(
                #     infiles,
                #     outfile,
                #     dataset_id,
                #     types=types,
                #     split=split,
                #     progress={"description": "Generating GeoJSON files..."},
                # )

        with console.status("Writing GeoPackage...") as status:
            df = gpd.GeoDataFrame()

            for filename in glob.glob(str(Path(tdir, "*.geojson"))):
                subdf = read_dataframe(filename)
                df = pd.concat([df, subdf])

            # TODO: Use more accurate CRS
            centroid = df.to_crs(3857).centroid.to_crs(4326)

            df["longitude"] = centroid.x
            df["latitude"] = centroid.y

            dest_path = Path(config.data_dir, f"{dataset_id}.gpkg")
            write_dataframe(df, dest_path, driver="GPKG")

            config.datasets[dataset_id]["gpkg"] = dest_path
            config.save()

        console.print("Writing GeoPackage... [green]Done")

        # click.echo(f"\nCreated: {dest_path}")



import re
import zipfile
from os import PathLike
from pathlib import Path

import geopandas as gpd
from pyogrio import read_dataframe

from plateaukit import defaults, generators, geocoding
from plateaukit.area import Area
from plateaukit.config import Config
from plateaukit.logger import logger

# from pyogrio import read_dataframe


class Dataset:
    """This class represents a dataset.

    Attributes:
        dataset_id: Dataset ID.
        gdf: The GeoDataFrame of the dataset.
    """

    dataset_id: str
    gdf: gpd.GeoDataFrame | None = None

    def __init__(self, dataset_id: str):
        """Initialize a dataset.

        Args:
            dataset_id: Dataset ID
        """
        self.dataset_id = dataset_id
        # config = Config()
        # gpkg_path = config.data[dataset_id].get("gpkg")
        # self.gdf = read_dataframe(gpkg_path)

    def __repr__(self):
        return f"Dataset({self.dataset_id})"

    def load_gdf(self):
        """Load the GeoDataFrame from the prebuilt dataset."""

        from pyogrio import read_dataframe

        config = Config()
        gpkg_path = config.datasets[self.dataset_id].get("gpkg")

        if gpkg_path:
            self.gdf = read_dataframe(gpkg_path)
        else:
            raise RuntimeError("Missing GeoPackage; Please prebuild the dataset first")

    def get_area(self):
        """Get the entire area of the dataset."""

        if self.dataset_id.endswith(".cloud"):
            raise Exception("Not supported for cloud datasets")

        if self.gdf is None:
            self.load_gdf()

        area_gdf = self.gdf

        return Area(area_gdf)

    def area_from_bbox(self, bbox: list[float] | None = None):
        """Get the specified area from the dataset.

        Args:
            bbox: Bounding box of the area of interest.
                   If not specified, the area of the entire dataset will be returned.
        """

        if self.dataset_id.endswith(".cloud"):
            # Load data from remote flatgeobuf
            remote_fgb = defaults.cloud_base_url + self.dataset_id.replace(
                ".cloud", ".fgb"
            )
            area_gdf = read_dataframe(
                remote_fgb,
                bbox=tuple(bbox),
            )
        else:
            if self.gdf is None:
                self.load_gdf()

            area_gdf = (
                self.gdf.cx[bbox[0] : bbox[2], bbox[1] : bbox[3]] if bbox else self.gdf
            )

        # TODO: Error handling when area_gdf is empty

        return Area(area_gdf)

    def area_from_polygons(self, polygons: list[gpd.GeoDataFrame]):
        """Get an area from the dataset by polygons.

        Args:
            polygon: Polygon of the area of interest.
        """

        if self.dataset_id.endswith(".cloud"):
            raise Exception("Not supported for cloud datasets")

        if self.gdf is None:
            self.load_gdf()

        area_gdf = self.gdf[self.gdf.geometry.intersects(polygons)]

        return Area(area_gdf)

    def area_from_points(
        self, points: list[list[float]], size: list[float] = [1000, 1000]
    ):
        """Get an area from the dataset by points.

        Args:
            points: List of points of the area of interest.
            size: Size of the area of interest in meters.
        """

        bbox = geocoding._get_bbox(points)
        bbox = geocoding._pad_bbox(bbox, size)

        return self.area_from_bbox(bbox)

    def area_from_address(self, address: str, size: list[float] = [1000, 1000]):
        """Get an area from the dataset by an address.

        Args:
            address: Address of the area of interest.
            size: Size of the area of interest in meters.
        """

        bbox = geocoding.bbox_from_address(address, min_size=size)

        return self.area_from_bbox(bbox)

    def area_from_postcode(self, postcode: str):
        """Get an area from the dataset by a postcode.

        Args:
            postcode: Postal code of the area of interest.
        """

        bbox = geocoding.bbox_from_postcode(postcode)

        return self.area_from_bbox(bbox)

    def area_from_landmark(self, landmark: str, min_size: list[float] = [1000, 1000]):
        """Get an area from the dataset by a landmark.

        Args:
            landmark: Landmark of the area of interest.
        """
        bbox = geocoding.bbox_from_landmark(landmark, min_size=min_size)

        return self.area_from_bbox(bbox)

    # def search_area(self, keyword: str):
    #     print(keyword)

    def to_geojson(
        self,
        outfile: str | PathLike,
        *,
        types: list[str] = ["bldg"],
        altitude: bool = True,
        include_type: bool = False,
        split: int = 1,
        **kwargs,
    ):
        """Generate GeoJSON from PLATEAU datasets.

        Args:
            outfile: Output file path.
            types: CityGML object types to include.
            split: Split the output into specified number of files.
            **kwargs: Keyword arguments for the generator.
        """
        if not self.dataset_id:
            raise Exception("Missing dataset_id")

        if not types:
            raise Exception("Missing object types")

        # NOTE: this is intentional but to be refactored in the future
        config = Config()
        record = config.datasets[self.dataset_id]

        if "citygml" not in record:
            raise Exception("Missing CityGML data")
        file_path = Path(record["citygml"])

        infiles = []

        # TODO: Refactor
        for type in types:
            # TODO: Fix
            pat = re.compile(rf".*udx\/{type}\/.*\.gml$")
            if zipfile.is_zipfile(file_path):
                with zipfile.ZipFile(file_path) as f:
                    namelist = f.namelist()
                    targets = list(filter(lambda x: pat.match(x), namelist))
                    # print(targets)
                    infiles += [str(Path("/", target)) for target in targets]
                    # f.extractall(tdir, members=targets)
                    # infiles += [
                    #     str(Path(tdir, Path(file_path).stem, "udx", type, "*.gml"))
                    # ]
            else:
                infiles += [str(Path(file_path, "udx", type, "*.gml"))]

        logger.debug([types, infiles, outfile])

        # Codelists
        codelist_infiles = []
        pat = re.compile(rf".*codelists\/.*\.xml$")
        with zipfile.ZipFile(file_path) as f:
            namelist = f.namelist()
            codelist_infiles = list(filter(lambda x: pat.match(x), namelist))
            codelist_infiles = [str(Path("/", target)) for target in codelist_infiles]

        generators.geojson.geojson_from_citygml(
            infiles,
            outfile,
            types=types,
            altitude=altitude,
            include_type=include_type,
            split=split,
            zipfile=file_path,
            codelist_infiles=codelist_infiles,
            **kwargs,
        )

    def to_cityjson(
        self,
        outfile: str | PathLike,
        types: list[str] = ["bldg"],
        object_types=None,  # TODO: Handle this
        lod=[1, 2],
        ground: bool = False,
        split: int = 1,
        **kwargs,
    ):
        """Generate CityJSON from PLATEAU datasets.

        Args:
            outfile: Output file path.
            types: CityGML feature types.
            split: Split the output into specified number of files.
            **kwargs: Keyword arguments for the generator.
        """
        params = {}

        # if precision:
        #     params["precision"] = precision

        if not self.dataset_id:
            raise Exception("Missing dataset_id")

        config = Config()
        record = config.datasets[self.dataset_id]

        if "citygml" not in record:
            raise Exception("Missing CityGML data")

        file_path = Path(record["citygml"])

        infiles = []

        for type in types:
            # TODO: fix
            pat = re.compile(rf".*udx\/{type}\/.*\.gml$")

            if zipfile.is_zipfile(file_path):
                with zipfile.ZipFile(file_path) as f:
                    namelist = f.namelist()
                    targets = list(filter(lambda x: pat.match(x), namelist))
                    infiles += [str(Path("/", target)) for target in targets]
            else:
                infiles += [str(Path(file_path, "udx", type, "*.gml"))]

            logger.debug([types, infiles, outfile])

        # Sort by a part of filename to group by areas; TODO: Update this
        # try:
        #     infiles = sorted(infiles, key=lambda x: Path(x).stem.split("_")[0])
        # except:
        #     infiles = sorted(infiles)
        infiles = sorted(infiles)

        # Codelists
        codelist_infiles = []
        pat = re.compile(rf".*codelists\/.*\.xml$")
        with zipfile.ZipFile(file_path) as f:
            namelist = f.namelist()
            codelist_infiles = list(filter(lambda x: pat.match(x), namelist))
            codelist_infiles = [str(Path("/", target)) for target in codelist_infiles]

        generators.simplecityjson.cityjson_from_citygml(
            infiles,
            outfile,
            split=split,
            zipfile=file_path,
            lod=lod,
            ground=ground,
            codelist_infiles=codelist_infiles,
        )


def load_dataset(dataset_id: str) -> Dataset:
    """Load a dataset.

    Args:
        dataset_id: Dataset ID.

    Returns:
        Dataset: Dataset object.
    """
    return Dataset(dataset_id)



from .city_gml_parser import CityGMLParser, PLATEAUCityGMLParser  # noqa
from .city_object_parser import CityObjectParser, PLATEAUCityObjectParser  # noqa
from .codelist_parser import CodelistParser  # noqa



import concurrent.futures
import io
import json
import math
import os.path
from decimal import Decimal
from multiprocessing import Manager
from os import PathLike
from pathlib import Path

from bidict import bidict
from fs import open_fs
from rich.progress import Progress

from plateaukit import parallel, utils
from plateaukit.logger import logger
from plateaukit.parsers import PLATEAUCityGMLParser
from plateaukit.utils import dict_key_to_camel_case

# from json_stream import streamable_dict


class VerticesMap:
    counter: int
    index_by_vertex: bidict

    def __init__(self):
        self.counter = 0
        self.index_by_vertex = bidict()

    def to_index(self, vertex):
        if vertex not in self.index_by_vertex:
            self.index_by_vertex[vertex] = self.counter
            self.counter += 1
        return self.index_by_vertex[vertex]

    # def to_vertex(self, index):
    #     return self.index_by_vertex.inverse[index]

    # def exists_index(self, index):
    #     return index in self.index_by_vertex.inverse

    @property
    def vertices(self):
        return list(self.index_by_vertex.keys())

    def __contains__(self, vertex):
        return vertex in self.index_by_vertex


def _get_nesting_level(l):
    if isinstance(l, list):
        return 1 + max([_get_nesting_level(x) for x in l])
    else:
        return 0


def _shift_to_ground(boundaries: list):
    level = _get_nesting_level(boundaries)

    # Nest level 3
    if level == 3:
        min_z = min(
            [
                min([min([point[2] for point in region]) for region in surface])
                for surface in boundaries
            ]
        )
        boundaries = [
            [
                [(point[0], point[1], point[2] - min_z) for point in region]
                for region in surface
            ]
            for surface in boundaries
        ]
    # Nest level 4
    elif level == 4:
        min_z = min(
            [
                min(
                    [
                        min([min([point[2] for point in shell]) for shell in solid])
                        for solid in surface
                    ]
                )
                for surface in boundaries
            ]
        )
        boundaries = [
            [
                [
                    [(point[0], point[1], point[2] - min_z) for point in shell]
                    for shell in solid
                ]
                for solid in surface
            ]
            for surface in boundaries
        ]

    return boundaries


def get_indexed_boundaries(geometry, vertices_map: VerticesMap, ground=False):
    # TODO: handling composite surface seriously
    # print("get_indexed_boundaries")
    # print("type", geometry["type"])

    boundaries = geometry["boundaries"]
    indexed_boundaries = []

    if geometry["type"] == "MultiPoint":
        for point in geometry["boundaries"]:
            index = vertices_map.to_index(point)
            indexed_boundaries.append(index)
        return indexed_boundaries, vertices_map

    elif geometry["type"] == "MultiLineString":
        for line in geometry["boundaries"]:
            indexed_line = []
            for point in line:
                index = vertices_map.to_index(point)
                indexed_line.append(index)
            indexed_boundaries.append(indexed_line)
        return indexed_boundaries, vertices_map

    elif geometry["type"] in ["MultiSurface", "CompositeSurface"]:
        if ground:
            boundaries = _shift_to_ground(geometry["boundaries"])
        for surface in boundaries:
            # print("surface", surface)
            indexed_surface = []
            for region in surface:
                # print("region", region)

                # NOTE: Reverse the order of vertices; this is necessary for
                # being rendered correctly in cityjson-threejs-loader
                # for SOME datasets
                # eg. plateau-13213-higashimurayama-shi-2020
                ## region = region[::-1]

                indexed_region = []
                unclosed_region = region[:-1]
                for point in unclosed_region:
                    index = vertices_map.to_index(point)
                    indexed_region.append(index)
                indexed_surface.append(indexed_region)
            indexed_boundaries.append(indexed_surface)
        # print("indexed_boundaries", indexed_boundaries, "aaa" * 100)
        return indexed_boundaries, vertices_map

    elif geometry["type"] == "Solid":
        if ground:
            boundaries = _shift_to_ground(geometry["boundaries"])
        for shell in boundaries:
            # print("shell", shell)
            indexed_shell = []
            for surface in shell:
                indexed_surface = []
                for region in surface:
                    # print("region", region)
                    indexed_region = []
                    unclosed_region = region[:-1]
                    for point in unclosed_region:
                        # print("point", point)
                        index = vertices_map.to_index(point)
                        indexed_region.append(index)
                    # print("indexed_region", indexed_region)
                    indexed_surface.append(indexed_region)
                indexed_shell.append(indexed_surface)
            indexed_boundaries.append(indexed_shell)
        # print("indexed_boundaries", indexed_boundaries)
        return indexed_boundaries, vertices_map
    else:
        raise NotImplementedError()

    # TODO: MultiSolid, CompositeSolid

    # for surface in exterior:
    #     unclosed_surface = surface[:-1]
    #     # print("unclosed_surface", unclosed_surface)
    #     single_exterior_surface_exterior = []

    #     for chunk in unclosed_surface:
    #         index = vertices_map.to_index(chunk)
    #         single_exterior_surface_exterior.append(index)

    #     indexed_surface = [single_exterior_surface_exterior]
    #     indexed_surfaces.append(indexed_surface)

    # return indexed_surfaces, vertices_map


class CityJSONConverter:
    def __init__(self, target_epsg):
        self.target_epsg = target_epsg

        self.vertices_map = VerticesMap()

    # @streamable_dict
    def generate_city_object(
        self,
        infiles: list[str],
        *,
        object_types: list[str] | None,
        lod: list[int],
        ground: bool = False,
        codelist_infiles: list[str] | None = None,
        zipfile: str | PathLike | None = None,
        task_id=None,
        quit=None,
        _progress=None,
    ):
        # vertices_map = VerticesMap()

        # Load codelists
        codelist_file_map = None

        if zipfile is not None:
            zip_fs = open_fs(f"zip://{zipfile}")

        if codelist_infiles and infiles:
            base_path = Path(infiles[0]).parent  # TODO: Fix this
            codelist_file_map = dict()

            for codelist_infile in codelist_infiles:
                if zipfile is not None:
                    with zip_fs.open(codelist_infile, "rb") as f:
                        relative_path = os.path.relpath(codelist_infile, base_path)
                        codelist_file_map[relative_path] = io.BytesIO(f.read())
                else:
                    with open(codelist_infile, "rb") as f:
                        relative_path = os.path.relpath(codelist_infile, base_path)
                        codelist_file_map[relative_path] = io.BytesIO(f.read())

        parser = PLATEAUCityGMLParser(
            target_epsg=self.target_epsg, codelist_file_map=codelist_file_map
        )

        total = len(infiles) + 1  # + 1 for geojson.dump

        for i, infile in enumerate(infiles):
            if task_id is not None and _progress is not None:
                _progress[task_id] = {"progress": i + 1, "total": total}

            if zipfile is not None:
                with zip_fs.open(infile, "r") as f:
                    citygml = parser.parse(f)
            else:
                with open(infile, "r") as f:
                    # print(f"infile: {infile}")

                    citygml = parser.parse(f)

            for city_obj in citygml.city_objects:
                if quit and quit.is_set():
                    return

                # print("city_obj", city_obj)

                if object_types is not None and city_obj.type not in object_types:
                    continue

                indexed_geoms = []

                for geom in city_obj.geometry:
                    if geom["lod"] not in lod:
                        continue

                    # print("geom", geom)
                    boundaries, vertices_map = get_indexed_boundaries(
                        geom,
                        self.vertices_map,
                        ground=ground,
                    )

                    indexed_geom = dict(
                        geom, lod=str(geom["lod"]), boundaries=boundaries
                    )

                    # print("indexed_geom", indexed_geom)
                    indexed_geoms.append(indexed_geom)

                    self.vertices_map = vertices_map

                # print("indexed_geoms", indexed_geoms)

                # obj_id = city_obj.attributes.get("building_id", city_obj.id)
                obj_id = city_obj.id

                yield obj_id, {
                    "type": city_obj.type,  # TODO: There is no Track type etc. in CityJSON
                    "attributes": dict_key_to_camel_case(city_obj.attributes),
                    # "attributes": {"建物ID": "13104-bldg-52530", "measuredHeight": 61.9},
                    # "children": [
                    #     "ID_22730c8f-9fbc-4d58-88dd-5569d7480fad",
                    #     "ID_598f2fab-030f-429c-b938-a222e04d8e4b",
                    #     "ID_db473977-e95e-4075-b0be-55eb65974610",
                    #     "ID_ac26b2cb-553e-428a-9f10-2659419e824d",
                    # ],
                    "geometry": indexed_geoms,
                    # [
                    # {
                    #     "type": "MultiSurface",
                    #     "lod": "0",
                    #     "boundaries": [],
                    # },
                    # {
                    #     "type": "Solid",
                    #     "lod": "1",
                    #     "boundaries": boundaries,
                    #     # "semantics": {
                    #     #     "surfaces": [],
                    #     #     "values": [],
                    #     # },
                    #     # "texture": {"rgbTexture": {"values": []}},
                    # },
                    # ],
                    # "address": [{"Country": "日本", "Locality": "東京都新宿区西新宿一丁目"}],
                }

        if zipfile is not None:
            zip_fs.close()


def cityson_from_gml_serial_with_quit(
    infiles,
    outfile,
    object_types,
    lod,
    ground,
    codelist_infiles,
    zipfile=None,
    task_id=None,
    quit=None,
    _progress=None,
    **opts,
):
    # TODO: logger does not work; must use QueueHandler
    logger.debug("[*] cityson_from_gml_serial_with_quit")

    target_epsg = 3857  # Web Mercator
    # target_epsg = 4326  # WGS84
    # target_epsg = 32654  # WGS84 / UTM zone 54N
    # target_epsg = 6677  # JGD2011 / Japan Plane Rectangular CS IV

    converter = CityJSONConverter(target_epsg=target_epsg)

    city_objects = dict(
        converter.generate_city_object(
            infiles,
            object_types=object_types,
            lod=lod,
            ground=ground,
            codelist_infiles=codelist_infiles,
            zipfile=zipfile,
            task_id=task_id,
            quit=quit,
            _progress=_progress,
        )
    )

    # print(city_objects)

    # vertices = [
    #     transformer.transform(*vertice) for vertice in converter.vertices_map.vertices
    # ]
    vertices = converter.vertices_map.vertices

    result = {
        "type": "CityJSON",
        "version": "2.0",
        "extensions": {},
        "transform": {"scale": [1.0, 1.0, 1.0], "translate": [0.0, 0.0, 0.0]},
        "metadata": {
            "referenceSystem": f"https://www.opengis.net/def/crs/EPSG/0/{target_epsg}",
        },
        "CityObjects": city_objects,
        "vertices": vertices,
        # "appearance": {},
        # "geometry-templates": {},
    }
    # result_debug = json.dumps(result, indent=2, ensure_ascii=False)
    # print(result_debug)

    with open(outfile, "w") as f:
        json.dump(result, f, ensure_ascii=False, separators=(",", ":"))

    total = len(infiles) + 1  # + 1 for geojson.dump

    # Complete the progress bar
    _progress[task_id] = {"progress": total, "total": total}


def cityjson_from_citygml(
    infiles,
    outfile,
    *,
    split: int = 1,
    zipfile=None,
    precision=16,
    object_types=None,
    lod: list[int] = [1, 2],
    ground: bool = False,
    codelist_infiles=None,
    progress={},
):
    logger.debug("[*] cityjson_from_citygml")
    # vertices_map = VerticesMap()
    # city_objects = {}

    group_size = math.ceil(len(infiles) / split)
    logger.debug(f"GMLs per GeoJSON: {group_size}")
    infile_groups = utils.chunker(infiles, group_size)

    with Progress() as rprogress:
        overall_progress_description = progress.get("description", "Processing...")
        overall_task_id = rprogress.add_task(overall_progress_description)

        with Manager() as manager:
            quit = manager.Event()
            _progress = manager.dict()

            with concurrent.futures.ProcessPoolExecutor(max_workers=None) as pool:
                futures = []
                futures_status = dict()
                for i, infile_group in enumerate(infile_groups):
                    stem = Path(outfile).stem

                    if split > 1:
                        group_outfile = Path(outfile).with_stem(f"{stem}.{i + 1}")
                    else:
                        group_outfile = outfile

                    task_id = rprogress.add_task(
                        f"[cyan]Progress #{i + 1}", total=len(infile_group)
                    )
                    # task_id = None

                    logger.debug(f"group_outfile: {group_outfile}")

                    future = pool.submit(
                        cityson_from_gml_serial_with_quit,
                        infile_group,
                        group_outfile,
                        object_types=object_types,
                        lod=lod,
                        ground=ground,
                        codelist_infiles=codelist_infiles,
                        zipfile=zipfile,
                        task_id=task_id,
                        _progress=_progress,
                        quit=quit,
                        # **kwargs,
                    )

                    futures.append(future)
                    futures_status[future] = {
                        "task_id": task_id,
                        "counter": i + 1,
                        "failed": False,
                    }

                parallel.wait_futures(
                    futures,
                    pool,
                    quit,
                    futures_status,
                    overall_progress={
                        "task_id": overall_task_id,
                        "description": overall_progress_description,
                    },
                    rich_progress=rprogress,
                    shared_progress_status=_progress,
                )



import re
import warnings

from plateaukit.constants import nsmap


def extract_gml_id(tree):
    path = f"./[@{nsmap['gml']}id]"
    result = tree.find(path)
    if result is None:
        warnings.warn(
            "gml:id not found"
            # f"gml:id not found\n{etree.tostring(tree, pretty_print=True).decode()}"
        )
        return None
    id = result.get(f"{nsmap['gml']}id")
    return id if id is not None else None


def extract_epsg(tree):
    path = f"./{nsmap['gml']}boundedBy/{nsmap['gml']}Envelope"
    result = tree.find(path)
    if result is None:
        raise Exception("EPSG not found")
    srs_name = result.get("srsName")
    m = re.match(r"http://www.opengis.net/def/crs/EPSG/0/(\d+)", srs_name)
    if not m:
        raise Exception("Failed to parse EPSG")
    return m.group(1)



import tempfile

from geopandas import GeoDataFrame


class Area:
    """This class represents an area of interest."""

    gdf: GeoDataFrame

    def __init__(self, gdf: GeoDataFrame) -> None:
        """Initialize an area from a GeoDataFrame.

        Args:
            gdf: GeoDataFrame of the area of interest.
        """
        self.gdf = gdf

    def __repr__(self) -> str:
        return f"Area()"

    def get_area(self, bbox: list[float] | None = None):
        """Get the specified area from the dataset.

        Args:
            bbox: Bounding box of the area of interest.
                   If not specified, the area of the entire dataset will be returned.
        """

        area_gdf = (
            self.gdf.cx[bbox[0] : bbox[2], bbox[1] : bbox[3]] if bbox else self.gdf
        )

        # TODO: Error handling when area_gdf is empty

        return Area(area_gdf)

    def get_centroid(self) -> list[float]:
        """Get the centroid of the area of interest."""

        bbox = self.gdf.total_bounds
        return [(bbox[0] + bbox[2]) / 2, (bbox[1] + bbox[3]) / 2]

    def show(self, embed: bool = True):
        import pydeck

        # cx, cy = self.get_centroid()

        bbox = self.gdf.total_bounds
        points = [(bbox[0], bbox[1]), (bbox[2], bbox[3])]
        # print(points)

        view_state = pydeck.data_utils.compute_view(points, view_proportion=1)
        view_state.pitch = 45

        deck = pydeck.Deck(
            layers=[
                # pydeck.Layer(
                #     "PolygonLayer",
                #     data=self.gdf,
                #     get_polygon="geometry.coordinates",
                #     filled=True,
                #     stroked=False,
                #     get_fill_color=[0, 0, 0, 20],
                #     get_line_color=[0, 0, 0, 20],
                # )
                pydeck.Layer(
                    "GeoJsonLayer",
                    data=self.gdf,
                    filled=True,
                    get_fill_color="fill_color || color || [255, 255, 255]",
                    # get_fill_color=[255, 255, 255, 240],
                    # get_line_color=[255, 255, 255],
                    extruded=True,
                    # wireframe=True,
                    get_elevation="measuredHeight",
                    pickable=True,
                )
            ],
            initial_view_state=view_state,
            # initial_view_state=pydeck.ViewState(
            #     latitude=cy,
            #     longitude=cx,
            #     zoom=1,
            #     pitch=0,
            #     bearing=0,
            # ),
            tooltip={
                "html": "<b>ID:</b> {buildingId}",
                "style": {
                    "font-family": "sans-serif",
                    "font-size": "8px",
                    "color": "white",
                },
            },
        )

        try:
            if __IPYTHON__:  # type: ignore
                return deck.to_html()
        except:
            pass

        with tempfile.NamedTemporaryFile(suffix=".html", delete=False) as fp:
            deck.to_html(fp.name, open_browser=True)

        # raise NotImplementedError()

    def dict(self):
        return self.gdf.to_dict()

    def to_geojson(self):
        """Convert the area to GeoJSON."""

        data = self.gdf.to_json()
        # print(data)
        return data

    def to_cityjson(self):
        """Convert the area to CityJSON."""

        raise NotImplementedError()

    @property
    def buildings(self):
        from plateaukit.models import Building

        # TODO: Filter and return buildings only

        for row in self.gdf.itertuples():
            df = self.gdf.loc[[row.Index]]

            yield Building(df)



from os import PathLike

from plateaukit.formats.citygml import CityGMLDataset


class CityGMLDatasetParser:
    def __init__(self):
        pass

    def load_dataset(self, infile: PathLike):
        return CityGMLDataset(infile)



import atexit
import importlib.resources
import platform
import subprocess
from contextlib import ExitStack
from pathlib import Path

import click

CITYGML_TOOLS_VERSION = "2.0.0"
CITYGML_TOOLS_DIR = f"external/citygml-tools-{CITYGML_TOOLS_VERSION}"


def _get_citygml_tools_path():
    file_manager = ExitStack()
    atexit.register(file_manager.close)
    filename = (
        "citygml-tools.bat" if platform.system() == "Windows" else "citygml-tools"
    )
    ref = Path(
        importlib.resources.files("plateaukit").parent, CITYGML_TOOLS_DIR, filename
    )
    path = file_manager.enter_context(importlib.resources.as_file(ref))
    return str(path)


def cityjson_from_citygml(infiles, vertex_precision=16):
    citygml_tools_path = _get_citygml_tools_path()
    args = [
        citygml_tools_path,
        "to-cityjson",
        f"--vertex-precision={vertex_precision}",
    ]
    args.extend(infiles)
    try:
        result = subprocess.run(
            args,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            stdin=subprocess.DEVNULL,
            check=True,
        )
    except subprocess.CalledProcessError as e:
        click.echo(
            "STDOUT from citygml-tools:\n{}\n".format(e.stdout.decode("utf-8")),
            err=True,
        )
        click.echo(
            "STDERR from citygml-tools:\n{}\n".format(e.stderr.decode("utf-8")),
            err=True,
        )
        raise
    except:
        raise



import click
from prettytable import PrettyTable
from rich.console import Console

from plateaukit import generators
from plateaukit.config import Config
from plateaukit.dataset import load_dataset
from plateaukit.installer import install_dataset, uninstall_dataset
from plateaukit.logger import logger, set_log_level
from plateaukit.prebuild import prebuild


def list_available_datasets(is_all=False):
    from plateaukit.download import city_list

    table = PrettyTable()
    table.field_names = ["id", "name", "version", "homepage"]
    table.add_row(["all", "(全都市)", "", ""])
    for city in city_list:
        if city.get("latest", False) or is_all:
            table.add_row(
                [
                    city["dataset_id"],
                    city["city_name"],
                    city["version"],
                    city["homepage"],
                ]
            )
    print(table)


# https://alexdelorenzo.dev/notes/click.html
class OrderCommands(click.Group):
    def list_commands(self, ctx):
        return list(self.commands)


@click.group(
    cls=OrderCommands,
    context_settings=dict(help_option_names=["-h", "--help"]),
    no_args_is_help=True,
)
@click.option(
    "-v", "--verbose", is_flag=True, default=False, help="Enable verbose mode."
)
def cli(verbose):
    if verbose:
        set_log_level("DEBUG")


@cli.command("list")
@click.option(
    "--local", is_flag=True, default=False, help="Show installed datasets only."
)
@click.option(
    "--all",
    is_flag=True,
    default=False,
    help="Show all versions of datasets including old ones.",
)
def list_cmd(local, all):
    """List available and installed PLATEAU datasets.

    利用可能およびインストール済みのデータセットを表示します。
    """
    from plateaukit.download import city_list

    config = Config()

    if local:
        table = PrettyTable()
        table.field_names = ["id", "name", "homepage", "formats"]
        for dataset_id, record in config.datasets.items():
            city = next(
                filter(lambda x: x["dataset_id"] == dataset_id, city_list), None
            )
            if not city:
                continue
            table.add_row(
                [
                    dataset_id,
                    city["city_name"],
                    city["homepage"],
                    " ".join(
                        [x for x in ["citygml", "3dtiles", "gpkg"] if x in record]
                    ),
                ]
            )
        print(table)
        return
    else:
        list_available_datasets(is_all=all)
        return


@cli.command("install")
@click.argument("dataset_id", nargs=1, required=False)
@click.option(
    "--format",
    type=click.Choice(["citygml", "3dtiles"], case_sensitive=False),
    default="citygml",
)
@click.option("--local", help="Install local file. (without copying)")
@click.option(
    "--prebuild", "run_prebuild", is_flag=True, default=True, help="Prebuild dataset."
)
@click.option("--force", is_flag=True, default=False, help="Force install.")
@click.option("--download-only", is_flag=True, default=False)
@click.option("-l", "--list", is_flag=True, help="List all latest available datasets.")
@click.option(
    "--list-all", is_flag=True, help="List all available datasets including old ones."
)
@click.option("-v", "is_verbose", is_flag=True, help="Verbose mode.")
def install_cmd(
    dataset_id,
    format,
    local,
    run_prebuild,
    force,
    download_only,
    list,
    list_all,
    is_verbose,
):
    """Download and install PLATEAU datasets.

    PLATEAU データセットをダウンロード・インストールします。
    """
    if not dataset_id and not (list or list_all):
        raise click.UsageError("Missing argument/option: dataset_id or -l/--list")

    if list or list_all:
        list_available_datasets(is_all=list_all)
        return

    if dataset_id:
        try:
            install_dataset(dataset_id, format, local, force)
        except Exception as e:
            raise click.UsageError(str(e))

    if run_prebuild:
        prebuild(dataset_id, simple_output=False if is_verbose else True)


@cli.command("uninstall")
@click.argument("dataset_id", nargs=1, required=False)
@click.option(
    "--format",
    "-f",
    "formats",
    type=click.Choice(
        ["citygml", "3dtiles", "gpkg"],
        case_sensitive=False,
    ),
    default=None,
    multiple=True,
)
@click.option("--keep-files", is_flag=True, default=False)
def uninstall_cmd(dataset_id, formats, keep_files):
    """Uninstall PLATEAU datasets.

    PLATEAU データセットをアンインストールします。
    """
    if not dataset_id:
        raise Exception("Missing argument")

    config = Config()
    formats = formats or config.datasets[dataset_id].keys()

    # TODO: Fix duplicated code in uninstall_dataset
    if not keep_files:
        paths = []
        for format in formats:
            path = config.datasets[dataset_id].get(format)
            if not path:
                raise RuntimeError(f"Missing files in record for '{format}'")
            paths.append(path)
        click.echo("Would remove:")
        for path in paths:
            click.echo(f"  {path}")
        if click.confirm("Proceed?"):
            uninstall_dataset(dataset_id, formats, keep_files=False)
    else:
        uninstall_dataset(dataset_id, formats, keep_files=True)


@cli.command("prebuild")
@click.argument("dataset_id", nargs=1, required=True)
@click.option(
    "--split", default=10, help="Split the output into specified number of files"
)
@click.option("-v", "is_verbose", is_flag=True, default=False, help="Verbose")
def prebuild_cmd(dataset_id, split, is_verbose):
    """Prebuild PLATEAU datasets.

    PLATEAU データセットを事前ビルドします。
    """

    prebuild(dataset_id, split=split, simple_output=False if is_verbose else True)


@cli.command("generate-cityjson")
@click.argument("infiles", nargs=-1)
@click.argument("outfile", nargs=1, required=True)
@click.option("--dataset", "dataset_id", help='Dataset ID (e.g. "plateau-tokyo23ku")')
@click.option(
    "--type",
    "-t",
    "types",
    type=click.Choice(
        ["bldg", "brid", "dem", "fld", "frn", "lsld", "luse", "tran", "urf"],
        case_sensitive=True,
    ),
    default=["bldg"],
    multiple=True,
)
@click.option(
    "--split", default=1, help="Split the output into specified number of files"
)
@click.option(
    "--ground", is_flag=True, default=False, help="Shift objects to the ground level"
)
# # @click.option(
# #     "--precision",
# #     help="Number of decimal places to keep for geometry vertices (default: 16).",
# )
def generate_cityjson(infiles, outfile, dataset_id, types, split, ground):
    """Generate CityJSON from PLATEAU datasets.

    PLATEAU データセットから CityJSON を生成します。
    """

    if not infiles and not dataset_id:
        raise click.UsageError("Missing argument/option: infiles or --dataset")

    if infiles and dataset_id:
        raise click.UsageError("Too many argument")

    params = {}

    # if precision:
    #     params["precision"] = precision

    # obj_types_map = {
    #     "bldg": ["Building"],
    #     "brid": ["Bridge"],
    #     "tran": ["Road"],
    #     # "dem": ["ReliefFeature"],
    #     # "fld": ["LandUse"],
    #     # "frn": ["WaterBody"],
    #     # "lsld": ["TransportationComplex"],
    #     # "luse": ["LandUse"],
    #     # "urf": ["CityFurniture"],
    # }

    # obj_types = sum([obj_types_map[type] for type in types], [])
    # logger.debug(obj_types)

    if dataset_id:
        dataset = load_dataset(dataset_id)
        dataset.to_cityjson(outfile, types=types, split=split, ground=ground, **params)

    else:
        generators.simplecityjson.cityjson_from_citygml(infiles, outfile, **params)


def _generate_geojson(
    infiles, outfile, dataset_id: str, types: list[str], split: int, **kwargs
):
    """Generate GeoJSON from PLATEAU datasets."""

    if not infiles and not dataset_id:
        raise click.UsageError("Missing argument: infiles or dataset")

    if infiles and dataset_id:
        raise click.UsageError("Too many arguments")

    if dataset_id:
        dataset = load_dataset(dataset_id)
        dataset.to_geojson(outfile, types=types, split=split, **kwargs)
    else:
        generators.geojson.geojson_from_citygml(
            infiles, outfile, types=types, split=split, **kwargs
        )


@cli.command("generate-geojson")
@click.argument("infiles", nargs=-1)
@click.argument("outfile", nargs=1, required=True)
@click.option("--dataset", "dataset_id", help='Dataset ID (e.g. "plateau-tokyo23ku")')
@click.option(
    "--type",
    "-t",
    "types",
    type=click.Choice(
        ["bldg", "brid", "dem", "fld", "frn", "lsld", "luse", "tran", "urf"],
        case_sensitive=True,
    ),
    default=["bldg"],
    multiple=True,
)
@click.option("--split", default=1)
def generate_geojson(infiles, outfile, dataset_id: str, types: list[str], split: int):
    """Generate GeoJSON from PLATEAU datasets.

    PLATEAU データセットから GeoJSON を生成します。
    """

    _generate_geojson(infiles, outfile, dataset_id, types=types, split=split)


@cli.command("generate-qmesh")
@click.argument("infiles", nargs=-1)
@click.argument("outfile", nargs=1, required=True)
def generate_qmesh(infiles, outfile):
    """Generate Quantized Mesh from PLATEAU datasets. (Requires `pip install 'plateaukit[quantized_mesh]'`)

    PLATEAU データセットから Quantized Mesh を生成します。(要 `pip install 'plateaukit[quantized_mesh]'`)
    """

    generators.triangles_from_gml(infiles)


# @cli.command("generate-gpkg")
# @click.argument("infiles", nargs=-1, required=True)
# @click.argument("outfile", nargs=1, required=True)
# def generate_gpkg(infiles, outfile):
#     """Generate GeoPackage from PLATEAU GeoJSON."""
#     expanded_infiles = []
#     for infile in infiles:
#         expanded_infiles.extend(glob.glob(infile))
#     generators.utils.geojson_to_gpkg(expanded_infiles, outfile)

# @cli.command("generate-heightmap")
# @click.argument("infiles", nargs=-1)
# @click.argument("outfile", nargs=1, required=True)
# def generate_heightmap(infiles, outfile):
#     """Generate GeoTIFF heightmap from PLATEAU CityGML."""
#     expanded_infiles = []
#     for infile in infiles:
#         expanded_infiles.extend(glob.glob(infile))
#     generators.triangles_from_gml(expanded_infiles)


# @cli.command("extract-properties")
# @click.argument("infiles", nargs=-1, required=True)
# @click.argument("outfile", nargs=1, required=True)
# def extract_properties(infiles, outfile):
#     """Extract properties from PLATEAU CityGML."""
#     expanded_infiles = []
#     for infile in infiles:
#         expanded_infiles.extend(glob.glob(infile))
#     run_async(extractors.commands.extract_properties(expanded_infiles, outfile))


@cli.command("config")
def info_cmd():
    """Show PlateauKit configuration information.

    PlateauKit の設定情報を表示します。
    """
    import importlib.metadata
    import json

    try:
        __version__ = importlib.metadata.version("plateaukit")
    except importlib.metadata.PackageNotFoundError:
        __version__ = "unknown"

    config = Config()
    click.echo(f"Version: {__version__}")
    click.echo(f"Config path: {config.path}")
    click.echo(f"Data directory: {config.data_dir}")
    click.echo(f"{json.dumps(config.datasets, indent=2, ensure_ascii=False)}")


@cli.command("info")
@click.argument("dataset_id", nargs=1, required=True)
def info_cmd(dataset_id):
    """Show PLATEAU dataset information.

    PLATEAU データセットの情報を表示します。
    """
    from plateaukit.download import city_list
    from plateaukit.formats.citygml import CityGMLDataset
    from plateaukit.parsers import constants

    config = Config()

    remote_info = next(filter(lambda x: x["dataset_id"] == dataset_id, city_list), None)

    if not remote_info:
        raise click.UsageError(f"Unknown dataset: {dataset_id}")

    console = Console(highlight=False)

    console.print(f"[b]{dataset_id}[/b]")
    console.print(f"[u]{remote_info['homepage']}")
    console.print(f"[b]Name:[/b] {remote_info['city_name']}")
    console.print(f"[b]Version:[/b] {remote_info['version']}")

    record = config.datasets.get(dataset_id, None)

    if not record:
        console.print(f"[b]Installed Files:[/b]\n  (None)")
        return

    console.print(f"[b]Installed Files:[/b]")
    for format, path in record.items():
        console.print(f"  {format}: {path}")

    dataset = CityGMLDataset(dataset_id)
    udx_dirs = dataset.udx_dirs()

    console.print(f"[b]Data Types:[/b]")
    for udx_type, path in udx_dirs.items():
        type_name = constants.data_type_display_names.get(udx_type, {}).get(
            "ja", udx_type
        )
        console.print(f"  {type_name} ({udx_type}): {path}")

    console.print(f"[b]Attributes:[/b]")

    with console.status("[bold]Processing...", spinner="simpleDotsScrolling") as status:
        for udx_type in udx_dirs.keys():
            type_name = constants.data_type_display_names.get(udx_type, {}).get(
                "ja", udx_type
            )
            console.print(f"  [b]{type_name} ({udx_type}):[/b]")
            for attr in dataset.scan_attributes(udx_type):
                console.print(f"    {attr.get('name')} ({attr['tag']})")


if __name__ == "__main__":
    cli()



from dataclasses import dataclass
from typing import Any

import pyproj

from plateaukit import extractors, utils
from plateaukit.parsers.constants import nsmap

MAPLIBRE_SCALE_FACTOR = 10000000
MERCATOR_HALF_WORLD_LENGTH = 20037508.342789243906736373901367187500

object_type_tags = {
    "Building": f"{{{nsmap['bldg']}}}Building",
    "Road": f"{{{nsmap['tran']}}}Road",
    "Bridge": f"{{{nsmap['brid']}}}Bridge",
}

geometry_type_tags = [
    {"type": "lod0RoofEdge", "tag": f"{{{nsmap['bldg']}}}lod0RoofEdge"},
    {"type": "lod0FootPrint", "tag": f"{{{nsmap['bldg']}}}lod0FootPrint"},
    {"type": "lod1Solid", "tag": f"{{{nsmap['bldg']}}}lod1Solid"},
    {"type": "lod2Solid", "tag": f"{{{nsmap['bldg']}}}lod2Solid"},
    {"type": "lod2MultiSurface", "tag": f"{{{nsmap['bldg']}}}lod2MultiSurface"},
    {"type": "lod1MultiSurface", "tag": f"{{{nsmap['tran']}}}lod1MultiSurface"},
    {"type": "lod2MultiSurface", "tag": f"{{{nsmap['brid']}}}lod2MultiSurface"},
]

# TODO: boundedBy


@dataclass
class CityObject:
    type: str
    id: str | None = None
    attributes: dict | None = None
    geometry: list[Any] | None = None
    # children: list[Self] | None = None
    # parent: Self | None = None


@dataclass
class Building(CityObject):
    address: dict | None = None


class GeometryParser:
    """A parser for CityGML geometries.

    Attributes:
        transformer: A pyproj.Transformer instance.
        maplibre: Whether to use MapLibre scale factor.
    """

    transformer: pyproj.Transformer
    maplibre: bool = False

    def __init__(self, transformer: pyproj.Transformer = None, maplibre: bool = False):
        self.transformer = transformer
        self.maplibre = maplibre

    def _transform(self, vertices, scale=(1, 1, 1), translate=(0, 0, 0)):
        # TODO: numpy

        transformed = []

        for vertex in vertices:
            x, y, z = vertex
            nx = (x * scale[0]) + translate[0]
            ny = (y * scale[1]) + translate[1]
            nz = (z * scale[2]) + translate[2]
            transformed.append((nx, ny, nz))

        return transformed

    def _adjust_mercator_to_maplibre(self, vertices):
        transformed = self._transform(
            vertices,
            scale=[-(0.5 * MAPLIBRE_SCALE_FACTOR) / MERCATOR_HALF_WORLD_LENGTH] * 2
            + [0.5 * MAPLIBRE_SCALE_FACTOR / MERCATOR_HALF_WORLD_LENGTH],
            translate=(
                0.5 * MAPLIBRE_SCALE_FACTOR,
                0.5 * MAPLIBRE_SCALE_FACTOR,
                0,
            ),
        )

        return transformed

    def extract_chunked_poslists(self, root):
        path = "/".join(
            [
                ".//gml:surfaceMember",  # TODO: Check this
                "gml:Polygon",
                "gml:exterior",
                "gml:LinearRing",
                "gml:posList",
            ]
        )
        # print(path)
        results = root.findall(path, nsmap)

        parsed = []

        for result in results:
            poslist = list(map(float, result.text.split(" ")))

            chunked = list(utils.chunker(poslist, 3))

            if self.maplibre and self.transformer:
                chunked = list(self.transformer.itransform(chunked))
                chunked = self._adjust_mercator_to_maplibre(chunked)
            elif self.transformer:
                chunked = list(self.transformer.itransform(chunked))

            surface = [chunked]
            parsed.append(surface)

        return parsed


class CityObjectParser:
    """A parser for CityGML objects.

    Attributes:
        transformer: A pyproj.Transformer instance.
        codelist_map: A map from codelist path to corresponding dict.
    """

    transformer: pyproj.Transformer
    codelist_map: dict[str, dict[str, str]] | None = {}

    def __init__(self, transformer: pyproj.Transformer = None, codelist_map={}):
        self.transformer = transformer
        self.codelist_map = codelist_map or {}


class PLATEAUCityObjectParser(CityObjectParser):
    # TODO: uro: attributes

    def _get_string_attribute(self, root, name):
        path = f"./gen:stringAttribute[@name='{name}']/gen:value"
        result = root.find(path, nsmap)
        return result.text if result is not None else None

    def _get_building_id(self, root):
        try:
            path = "./uro:buildingIDAttribute/uro:BuildingIDAttribute/uro:buildingID"
            result = root.find(path, nsmap)
            value = result.text if result is not None else None
            assert value is not None
            return value
        except:
            pass

        try:
            value = self._get_string_attribute(root, name="建物ID")
            assert value is not None
            return value
        except:
            pass

        return None

    def _get_measured_height(self, root):
        result = root.find(f"./bldg:measuredHeight", nsmap)
        value = result.text if result is not None else None
        value = float(value) if value is not None else None
        return value

    def _get_year_of_construction(self, root):
        result = root.find(f"./bldg:yearOfConstruction", nsmap)
        value = result.text if result is not None else None
        value = int(value) if value is not None else None
        return value

    def _get_storeys_above_ground(self, root):
        result = root.find(f"./bldg:storeysAboveGround", nsmap)
        value = result.text if result is not None else None
        value = int(value) if value is not None else None
        return value

    def _get_storeys_below_ground(self, root):
        result = root.find(f"./bldg:storeysBelowGround", nsmap)
        value = result.text if result is not None else None
        value = int(value) if value is not None else None
        return value

    def __get_codespace_attribute(self, root, xpath) -> str | None:
        el = root.find(xpath, nsmap)

        if el is None:
            return None

        # Check if el has codeSpace attribute
        if (code_space_path := el.get("codeSpace")) is not None:
            code_dict = self.codelist_map.get(code_space_path, None)

            if code_dict is None:
                return None

            key = el.text
            value = code_dict.get(key, None)
            return value

        value = el.text if el is not None else None
        return value

    def _get_name(self, root):
        value = self.__get_codespace_attribute(root, "./gml:name")
        return value

    def _get_usage(self, root):
        value = self.__get_codespace_attribute(root, "./bldg:usage")
        return value

    def _get_address(self, root):
        el = root.find(f"./bldg:address", nsmap)
        locality_name = el.find(f".//xAL:LocalityName", nsmap).text

        addr = {
            "locality_name": locality_name,
        }

        # print(addr)

        return addr

    def _get_geometry(self, root):
        geoms = []

        parser = GeometryParser(transformer=self.transformer)

        for type_tag in geometry_type_tags:
            type = type_tag["type"]
            tag = type_tag["tag"]

            el = root.find(f"./{tag}", nsmap)

            if el is None:
                continue

            if type in ["lod0RoofEdge", "lod0FootPrint"]:
                chunked_poslists = parser.extract_chunked_poslists(el)

                geom = {
                    "type": "MultiSurface",
                    "lod": 0,
                    "boundaries": chunked_poslists,
                    "semantics": {
                        "surfaces": [
                            {
                                "type": f"+{type}",
                            }
                        ],
                        "values": [0 for _ in range(len(chunked_poslists))],
                    },
                }

                geoms.append(geom)
            elif type in ["lod1MultiSurface"]:
                chunked_poslists = parser.extract_chunked_poslists(el)

                geom = {
                    "type": "MultiSurface",
                    "lod": 1,
                    "boundaries": chunked_poslists,
                    "semantics": {
                        "surfaces": [
                            {
                                "type": f"+{type}",
                            }
                        ],
                        "values": [0 for _ in range(len(chunked_poslists))],
                    },
                }

                geoms.append(geom)
            elif type in ["lod1Solid"]:
                # TODO: Fix this
                chunked_poslists = parser.extract_chunked_poslists(el)
                solid_boundaries = [chunked_poslists]

                geom = {
                    "type": "Solid",
                    "lod": 1,
                    "boundaries": solid_boundaries,
                    "semantics": {
                        "surfaces": [
                            {
                                "type": f"+{type}",
                            }
                        ],
                        "values": [0 for _ in range(len(solid_boundaries))],
                    },
                }

                geoms.append(geom)
            elif type in ["lod2MultiSurface"]:
                chunked_poslists = parser.extract_chunked_poslists(el)

                geom = {
                    "type": "MultiSurface",
                    "lod": 2,
                    "boundaries": chunked_poslists,
                    "semantics": {
                        "surfaces": [
                            {
                                "type": f"+{type}",
                            }
                        ],
                        "values": [0 for _ in range(len(chunked_poslists))],
                    },
                }

                geoms.append(geom)

        # Look through boundedBy
        bound_els = list(root.iterfind("./bldg:boundedBy", nsmap))
        for bound_el in bound_els:
            # TODO: Check semantics
            for type_tag in geometry_type_tags:
                type = type_tag["type"]
                tag = type_tag["tag"]

                el = bound_el.find(f".//{tag}", nsmap)

                if el is None:
                    continue

                if type in ["lod2MultiSurface"]:
                    chunked_poslists = parser.extract_chunked_poslists(el)

                    geom = {
                        "type": "MultiSurface",
                        "lod": 2,
                        "boundaries": chunked_poslists,
                        "semantics": {
                            "surfaces": [
                                {
                                    "type": f"+{type}",
                                }
                            ],
                            "values": [0 for _ in range(len(chunked_poslists))],
                        },
                    }

                    geoms.append(geom)

                else:
                    pass
                    # raise NotImplementedError()

        return geoms

    def parse(self, el):
        citygml_id = extractors.utils.extract_gml_id(el)
        address = None  # self._get_address(el)

        attributes = dict()

        if el.tag == object_type_tags["Building"]:
            attributes["building_id"] = self._get_building_id(el)

            # Optional attributes
            optional_attributes = {
                "measured_height": self._get_measured_height(el),
                "year_of_construction": self._get_year_of_construction(el),
                "storeys_above_ground": self._get_storeys_above_ground(el),
                "storeys_below_ground": self._get_storeys_below_ground(el),
                "name": self._get_name(el),
                "usage": self._get_usage(el),
            }
            optional_attributes = {
                k: v for k, v in optional_attributes.items() if v is not None
            }
            attributes.update(optional_attributes)

            geometry = self._get_geometry(el)

            obj = Building(
                type="Building",
                id=citygml_id,
                attributes=attributes,
                geometry=geometry,
                address=address,
            )
        elif el.tag == object_type_tags["Road"]:
            geometry = self._get_geometry(el)

            obj = CityObject(
                type="Road",
                id=citygml_id,
                attributes=attributes,
                geometry=geometry,
            )
        elif el.tag == object_type_tags["Bridge"]:
            geometry = self._get_geometry(el)

            obj = CityObject(
                type="Bridge",
                id=citygml_id,
                attributes=attributes,
                geometry=geometry,
            )
        else:
            raise NotImplementedError(f"Unknown object type: {el.tag}")

        return obj



# %%

from dataclasses import dataclass
from typing import Any, BinaryIO

import pyproj
from lxml import etree

from plateaukit import extractors
from plateaukit.parsers.city_object_parser import CityObject, PLATEAUCityObjectParser
from plateaukit.parsers.codelist_parser import CodelistParser
from plateaukit.parsers.constants import nsmap


@dataclass
class CityGML:
    city_objects: list[CityObject]


class CityGMLParser:
    pass


class PLATEAUCityGMLParser(CityGMLParser):
    """A parser for PLATEAU CityGML.

    Attributes:
        target_epsg: Target EPSG code.
        codelist_file_map: A map from codelist path to file object.
    """

    target_epsg: int
    codelist_file_map: dict[str, BinaryIO] | None = None

    def __init__(self, target_epsg: int = 4326, codelist_file_map=None):
        self.target_epsg = target_epsg
        self.codelist_file_map = codelist_file_map

    def parse(self, infile):
        tree = etree.parse(infile)
        root = tree.getroot()

        src_epsg = extractors.utils.extract_epsg(tree)  # 6697

        # TODO: Accept options like always_xy
        transformer = pyproj.Transformer.from_crs(src_epsg, self.target_epsg)

        # Parse codelists
        codelist_map = None

        if self.codelist_file_map:
            codelist_map = dict()

            parser = CodelistParser()

            for path, file_obj in self.codelist_file_map.items():
                codelist = parser.parse(file_obj)
                codelist_map[path] = codelist

        co_parser = PLATEAUCityObjectParser(
            transformer=transformer, codelist_map=codelist_map
        )

        objects = []

        for i, el in enumerate(root.iterfind(f"./core:cityObjectMember/*", nsmap)):
            obj = co_parser.parse(el)

            # print(obj)

            objects.append(obj)

        return CityGML(
            city_objects=objects,
        )

    # def city_objects(self):
    #     pass
